package mccity.heroes.skills.turret;

import com.herocraftonline.heroes.characters.Hero;
import com.herocraftonline.heroes.characters.party.HeroParty;
import com.herocraftonline.heroes.characters.skill.SkillConfigManager;
import mccity.heroes.skills.turret.integration.PluginsIntegration;
import net.minecraft.server.EntityArrow;
import net.minecraft.server.MathHelper;
import net.minecraft.server.WorldServer;
import org.bukkit.*;
import org.bukkit.block.Block;
import org.bukkit.block.BlockState;
import org.bukkit.block.Dispenser;
import org.bukkit.craftbukkit.CraftWorld;
import org.bukkit.entity.*;

import java.util.*;

public class Turret {

    private static final float Y_OFFSET_OVER_BASE_BLOCK = 1.6f;
    private static final float SCAN_RADIUS = 25;
    private static final int ATTACK_TARGET_FOR_TICKS = 80;
    private static final int SCAN_EVERY_TICKS = 30;

    private final String ownerName;
    private Hero owner; // null if owner is offline
    private final Block baseBlock;
    private final Location loc;
    private final World world;

    private LivingEntity target = null;
    private long startAttackCurrentTargetTick = 0;
    private long lastScanTick = 0;

    private float damage;
    private int knockback;
    private boolean fireArrow;
    private int attackPeriodTicks;
    private long lastAttackTick = 0;

    private final boolean unlimitedAmmo;
    private int ammo;
    private int ammoPerArrow;

    private final boolean persistent;
    private long endLifeTick;
    private boolean destroyOnOwnerLogout;
    private boolean destroyOnChunkUnload;

    public Turret(Hero owner, Block base, SkillTurret skill) {
        ownerName = owner.getPlayer().getName();
        this.owner = owner;
        this.baseBlock = base;
        this.loc = base.getLocation().add(0.5, Y_OFFSET_OVER_BASE_BLOCK, 0.5);
        this.world = loc.getWorld();

        attackPeriodTicks = SkillConfigManager.getUseSetting(owner, skill, "power.attack-period-ticks", 20, true);
        this.damage = SkillConfigManager.getUseSetting(owner, skill, "power.damage", 3, true);
        this.knockback = SkillConfigManager.getUseSetting(owner, skill, "power.knockback", 0, false);
        this.fireArrow = SkillConfigManager.getUseSetting(owner, skill, "power.fire-arrow", false);
        unlimitedAmmo = skill.isUnlimitedAmmoFor(owner);
        if (!unlimitedAmmo) {
            ammo = skill.getInitialAmmoFor(owner);
            ammoPerArrow = skill.getAmmoPerArrowFor(owner);
        }

        persistent = skill.isPersistentFor(owner);
        if (!persistent) {
            endLifeTick = world.getFullTime() + skill.getTurretLifeTimeFor(owner) / 50;
            destroyOnOwnerLogout = SkillConfigManager.getUseSetting(owner, skill, "lifetime.destroy-on-owner-logout", true);
            destroyOnChunkUnload = SkillConfigManager.getUseSetting(owner, skill, "lifetime.destroy-on-chunk-unload", true);
        }
    }

    public Turret(Map<?, ?> turretData) {
        ownerName = (String) turretData.get("owner-name");
        owner = null;

        Location baseBlockLoc = Utils.deserializeLocation((Map<?, ?>) turretData.get("base-block-loc"));
        loc = baseBlockLoc.clone().add(0.5, Y_OFFSET_OVER_BASE_BLOCK, 0.5);
        world = baseBlockLoc.getWorld();
        baseBlock = world.getBlockAt(baseBlockLoc);

        attackPeriodTicks = ((Number) turretData.get("attack-period-ticks")).intValue();
        damage = ((Number) turretData.get("damage")).floatValue();
        knockback = ((Number) turretData.get("knockback")).intValue();
        fireArrow = (Boolean) turretData.get("fire-arrow");

        unlimitedAmmo = (Boolean) turretData.get("unlimited-ammo");
        if (!unlimitedAmmo) {
            ammo = ((Number) turretData.get("ammo")).intValue();
            ammoPerArrow = ((Number) turretData.get("ammo-per-arrow")).intValue();
        }

        persistent = true;
    }

    public Map<String, Object> getDataEntry() {
        if (!persistent) return null;

        Map<String, Object> dataEntry = new LinkedHashMap<String, Object>();

        dataEntry.put("owner-name", ownerName);
        dataEntry.put("base-block-loc", Utils.serializeLocation(baseBlock.getLocation()));

        dataEntry.put("attack-period-ticks", attackPeriodTicks);
        dataEntry.put("damage", damage);
        dataEntry.put("knockback", knockback);
        dataEntry.put("fire-arrow", fireArrow);

        dataEntry.put("unlimited-ammo", unlimitedAmmo);
        if (!unlimitedAmmo) {
            dataEntry.put("ammo", ammo);
            dataEntry.put("ammo-per-arrow", ammoPerArrow);
        }

        return dataEntry;
    }

    public String getOwnerName() {
        return ownerName;
    }

    public void onAiTick(ExpirableEntityManager eyeManager) {
        long curTick = world.getFullTime();

        if (!unlimitedAmmo && ammo <= 0) {
            if (!tryLoadAmmo()) {
                world.playEffect(loc, Effect.SMOKE, 4);
                return;
            }
        }

        if (target != null && (target.isDead() || curTick >= startAttackCurrentTargetTick + ATTACK_TARGET_FOR_TICKS)) {
            target = null;
        }
        if (target == null && curTick >= lastScanTick + SCAN_EVERY_TICKS) {
            findNewTarget(eyeManager);
            lastScanTick = curTick;
            if (target != null) {
                startAttackCurrentTargetTick = curTick;
            }
        }

        if (target != null && (unlimitedAmmo || ammo > 0) && curTick >= lastAttackTick + attackPeriodTicks) {
            shootTarget();
            lastAttackTick = curTick;
            if (!unlimitedAmmo) {
                ammo--;
                if (ammo == 0) {
                    if (!tryLoadAmmo()) {
                        Utils.safeSend(ownerName, Messages.outOfAmmo);
                    }
                }
            }
        }
    }

    private boolean tryLoadAmmo() {
        BlockState baseState = baseBlock.getState();
        if (baseState instanceof Dispenser) {
            Dispenser dispenser = (Dispenser) baseState;
            if (Utils.tryConsumeItem(dispenser.getInventory(), Material.ARROW)) {
                ammo += ammoPerArrow;
                return true;
            }
        }
        return false;
    }

    private void findNewTarget(ExpirableEntityManager eyeManager) {
        EnderPearl testEntity = eyeManager.spawn(loc, EnderPearl.class, 10);
        List<Entity> nearEntities = testEntity.getNearbyEntities(SCAN_RADIUS, SCAN_RADIUS, SCAN_RADIUS);

        List<LivingEntity> possiblyTargets = new ArrayList<LivingEntity>();
        Set<String> friendlyPlayerNames = new HashSet<String>();
        for (Entity nearEntity : nearEntities) {
            if (!(nearEntity instanceof LivingEntity)) continue;
            LivingEntity nearLiving = (LivingEntity) nearEntity;

            if (nearLiving.getHealth() <= 0) continue; // dead

            if (nearLiving instanceof Player) {
                Player nearPlayer = (Player) nearLiving;
                if (nearPlayer.getName().equals(ownerName)) { // owner
                    friendlyPlayerNames.add(nearPlayer.getName());
                    continue;
                }
                if (nearPlayer.getGameMode() == GameMode.CREATIVE) { // creative
                    friendlyPlayerNames.add(nearPlayer.getName());
                    continue;
                }

                if (owner != null) { // owner is online
                    HeroParty party = owner.getParty();
                    if (party != null && party.isPartyMember(nearPlayer)) { // party
                        friendlyPlayerNames.add(nearPlayer.getName());
                        continue;
                    }
                }

                if (PluginsIntegration.instance.isFriendly(ownerName, nearPlayer)) {
                    friendlyPlayerNames.add(nearPlayer.getName());
                    continue;
                }
            }

            if (!Utils.isInLineOfSight(loc, nearLiving.getLocation().add(0, nearLiving.getEyeHeight(), 0))) continue;

            possiblyTargets.add(nearLiving);
        }

        target = selectTarget(possiblyTargets, friendlyPlayerNames);
    }

    private LivingEntity selectTarget(List<LivingEntity> targets, Set<String> friendlyPlayerNames) {
        if (targets.isEmpty()) return null;

        int targetIndex = -1;
        int maxPriority = 0;

        for (int i = 0; i < targets.size(); i++) {
            LivingEntity target = targets.get(i);
            int curPriority = 0;

            // skip friend's pets
            if (target instanceof Tameable) {
                Tameable pet = (Tameable) target;
                if (pet.getOwner() != null && friendlyPlayerNames.contains(pet.getOwner().getName())) {
                    continue;
                }
            }

            // distance modifier
            double distanceMod = loc.distance(target.getLocation());
            distanceMod = Math.min(Math.max(distanceMod, 1), 24);
            curPriority += (25 - (int) distanceMod);

            // player modifier
            if (target instanceof Player) {
                Player targetPlayer = (Player) target;
                curPriority += 20;
                if (targetPlayer.isSneaking()) curPriority -= 27; // sneaking
                if (owner != null && owner.isInCombatWith(targetPlayer)) curPriority += 30; // enemy!
            }

            if (curPriority > maxPriority) {
                maxPriority = curPriority;
                targetIndex = i;
            }
        }

        return (targetIndex == -1) ? null : targets.get(targetIndex);
    }

    private void shootTarget() {
        WorldServer mcWorld = ((CraftWorld) loc.getWorld()).getHandle();

        EntityArrow mcArrow = new EntityArrow(mcWorld);

        // reset fields setted in the EntityArrow constructor
        mcArrow.width = 0.6F;
        mcArrow.size = net.minecraft.server.Entity.EntitySize.SIZE_2;
        mcArrow.length = 1.8F;

        mcArrow.shooter = null;
        mcArrow.fromPlayer = false;
        Utils.setArrowDamage(mcArrow, damage);
        Utils.setArrowKnockback(mcArrow, knockback);
        if (fireArrow) {
            mcArrow.setOnFire(100);
        }

        // Based on Minecraft aim code (class EntityArrow)
        mcArrow.locY = loc.getY();
        double vecX = target.getLocation().getX() - loc.getX();
        double vecY = target.getLocation().getY() + target.getEyeHeight() - 0.7 - mcArrow.locY;
        double vecZ = target.getLocation().getZ() - loc.getZ();
        double vecXZ = (double) MathHelper.sqrt(vecX * vecX + vecZ * vecZ);
        float yaw = (float) (Math.atan2(vecZ, vecX) * 180.0 / Math.PI) - 90f;
        float pitch = (float) (-(Math.atan2(vecY, vecXZ) * 180.0 / Math.PI));
        double d4 = vecX / vecXZ;
        double d5 = vecZ / vecXZ;
        mcArrow.setPositionRotation(loc.getX() + d4, mcArrow.locY, loc.getZ() + d5, yaw, pitch);
        mcArrow.height = 0f;

        float f4 = (float) vecXZ * 0.2f;
        mcArrow.shoot(vecX, vecY + f4, vecZ, 1.6f, 12f);

        world.playEffect(loc, Effect.BOW_FIRE, 0);

        mcWorld.addEntity(mcArrow);
    }

    public void ownerJoined(Hero owner) {
        this.owner = owner;
    }

    public void ownerQuited() {
        this.owner = null;
    }

    public boolean isAlive() {
        if (persistent) {
            return !(baseBlock.getChunk().isLoaded() && baseBlock.getType() != Material.DISPENSER);
        } else {
            boolean alive = world.getFullTime() < endLifeTick;
            if (baseBlock.getChunk().isLoaded()) {
                alive = alive && baseBlock.getType() == Material.DISPENSER;
            } else {
                alive = alive && !destroyOnChunkUnload;
            }
            if (owner != null) {
                alive = alive && owner.canUseSkill("Turret");
            } else {
                alive = alive && !destroyOnOwnerLogout;
            }
            return alive;
        }
    }

    public Block getBaseBlock() {
        return baseBlock;
    }
}
